.csstage {
}

.csstage .unlockbox {
    color: #eeeeee;
    font-size: 20px;
    text-shadow: 1px 1px alpha(black, 0.8);
}

.csstage .clock {
    color: #eeeeee;
    text-shadow: 1px 1px alpha(black, 0.8);
}

.csstage .toppanel {
    border-color: alpha(white, .2);
    border-style: solid;
    color: white;
    background-color: transparent;
    background-image: linear-gradient(to bottom, rgba(0, 0, 0, .2), rgba(0, 0, 0, 0));
}

.csstage .audiopanel {
    border-width: 0 1px 1px 0;
    border-radius: 0 0 3px 0;
    background-color: rgba(255, 255, 255, .05);
}

.csstage .infopanel {
    border-width: 0 0 1px 1px;
    border-radius: 0 0 0 3px;
    background-color: rgba(255, 255, 255, .05);
}

.csstage .notificationwidget,
.csstage .powerwidget {
    font-size: 12px;
    font-weight: bold;
    color: white;
    background-color: transparent;
    background-image: none;
    text-shadow: 1px 1px alpha(black, 0.8);
    padding: 6px;
}

.csstage .auth-message {
    font-size: 15px;
    color: red;
}

.csstage .caps-message {
    font-size: 15px;
    color: orange;
}

.csstage .framedimage {
    border-radius: 4px;
    border: 4px solid;
    background-color: alpha(grey, .25);
    background-clip: border-box;
    border-color: @theme_selected_bg_color;
    box-shadow: 1px 1px alpha(black, 0.8);
}

.csstage .passwordentry {
    font-size: 15px;
    box-shadow: none;
    border-color: alpha(white, .2);
    border-style: solid;
    color: white;
    background-image: none;
    background-color: transparent;
    border-width: 1px;
    box-shadow: none;
}

.csstage .passwordentry progress {
    margin: 2px;
    padding: 0;
    border-radius: 0;
    border-width: 0 0 2px;
    border-color: alpha(@theme_selected_bg_color, .7);
}

.csstage .passwordentry progress:focus {
    background-color: transparent;
}

.csstage .passwordentry image.left {
    padding-right: 10px;
}

/* Use :backdrop for alt-text keyboard layout */
.csstage .passwordentry:backdrop {
    font-family: monospace;
    font-size: 14px;
    color: @theme_selected_bg_color;
}

.csstage .transparentbutton {
    -gtk-icon-style: requested;
    border-color: alpha(white, .2);
    border-image: none;
    box-shadow: none;
    background-image: none;
    background-color: transparent;
    border-radius: 20px;
    -gtk-outline-radius: 20px;
    outline-color: transparent;
    color: alpha(white, .7);
    border-width: 1px;
    padding: 4px;
}

.csstage .transparentbutton image {
    color: alpha(white, .7);
}

.csstage .transparentbutton:disabled image {
    color: #333333;
}

.csstage .passwordentry:focus,
.csstage .transparentbutton:focus {
    background-color: alpha(grey, .3);
    border-color: @theme_selected_bg_color;
    box-shadow: 1px 1px alpha(white, 0.1);
}

.csstage .passwordentry:hover,
.csstage .transparentbutton:hover {
    border-color: alpha(white, .2);
    background-color: alpha(grey, .2);
    box-shadow: 1px 1px alpha(white, 0.1);
}

.csstage .passwordentry:hover:focus,
.csstage .transparentbutton:hover:focus {
    border-color: @theme_selected_bg_color;
    background-color: alpha(grey, .4);
    box-shadow: 1px 1px alpha(white, 0.1);
}

.csstage .passwordentry:active,
.csstage .transparentbutton:active {
    border-color: @theme_selected_bg_color;
    background-color: alpha(grey, .6);
    box-shadow: 1px 1px alpha(white, 0.1);
}

.csstage .passwordentry:active:focus,
.csstage .transparentbutton:active:focus {
    border-color: @theme_selected_bg_color;
    background-color: alpha(grey, .9);
    box-shadow: 1px 1px alpha(white, 0.1);
}

.csstage .volumeslider {
    min-height: 24px;
    min-width: 100px;
    background-color: rgba(255, 255, 255, .1);
    color: @theme_selected_bg_color;
    padding: 3px 0px 3px 0px;
}

.csstage scale.volumeslider slider {
    min-height: 24px;
    min-width: 4px;
    margin: -9px; }

.csstage .volumeslider:disabled {
    background-color: alpha(white, .5);
}

.csstage .trackname {
    font-family: monospace;
    font-size: 12px;
    text-shadow: 1px 1px alpha(black, 0.8);
    background-image: none;
    background-color: transparent;
}

.csstage .albumartist {
    font-family: monospace;
    font-size: 8px;
    text-shadow: 1px 1px alpha(black, 0.8);
    background-image: none;
    background-color: transparent;
}

.csstage GtkViewport {
    background-color: transparent;
    background-image: none;
}
